'''Regex based parsing for sip uris.'''
import re
import multidict


# optionally extract headers that appear after
# a <> delimited contact, discarding the space
# between them if it exists
post_params_re = re.compile(r'(?P<the_rest>[^\>]*\>)'
                            r' ?(?P<params>.*)\Z')

# optionally extract either a quoted or unquoted
# contact name and extract the rest discarding the <>
# brackets if present
contact_re = re.compile(r'("(?P<quoted>[^"]+)" '
                        r'|(?P<unquoted>[^\<"]+) )?'
                        r'\<(?P<the_rest>.*)\>\Z')

# extract a scheme, discard the : and extract the rest
scheme_re = re.compile(r'(?P<scheme>[^:]+):(?P<the_rest>.*)')

# optionally extract a user part (either user
# or user:pass) and discard the @ symbol
user_part_re = re.compile(r'(?P<user_part>[^:@]+(:[^@:]+)?)@(?P<the_rest>.*)')

# extract a host component that may be [...]
# delimited in order to support ipv6 hosts
host_re = re.compile(r'(?P<host>('
                     r'\[[^\]]+])|([^:;?&]+)'
                     r')(?P<the_rest>.*)')

# optionally discard a : and extract a port number
port_re = re.compile(r':(?P<port>[^;?]*)(?P<the_rest>.*)')

# optionally discard a ; and extract a single key=val pair
param_re = re.compile(r';(?P<key>[^=]+)=(?P<val>[^;?]*)(?P<the_rest>.*)')

# optionally extract a key=val pair
base_header_re = r'(?P<key>[^=]+)=(?P<val>[^&]*)(?P<the_rest>.*)'
# match a ? then a key=val pair - only first header starts with ?
first_header_re = re.compile(rf'\?{base_header_re}')
# match a & then a key=val pair - matches (n>1)th headers
nth_header_re = re.compile(rf'&{base_header_re}')


def parse_post_params(uri):
    '''Attempt to match on headers after
    a contact+uri combination.

    Ex: `"User" <sip:localhost"> ;tag=abc`
        should extract the `;tag=abc`
    '''
    match = post_params_re.match(uri)
    if not match:
        return None, uri

    groups = match.groupdict()
    return groups['params'], groups['the_rest']


def parse_contact(uri):
    '''Attempt a contact match, and remove the
    proper sip uri from a <> pair if present
    '''
    match = contact_re.match(uri)
    if not match:
        return None, uri

    groups = match.groupdict()
    contact = groups.get('quoted')
    if contact is None:
        contact = groups.get('unquoted')

    return contact, groups['the_rest']


def parse_scheme(uri):
    '''Require a scheme match.'''
    match = scheme_re.match(uri)
    if match is None:
        raise ValueError('no scheme specified')
    groups = match.groupdict()

    scheme = groups['scheme'].lower()
    if scheme not in ('sip', 'sips'):
        raise ValueError(f'invalid scheme {scheme}')

    return scheme, groups['the_rest']


def parse_user_part(uri):
    '''Attempt to match user@ or user:pass@.'''
    match = user_part_re.match(uri)
    if match is None:
        return None, None, uri
    groups = match.groupdict()

    subparts = groups['user_part'].split(':')
    if len(subparts) == 2:
        user, password = subparts
    else:
        user, password = subparts[0], None

    return user, password, groups['the_rest']


def parse_host(uri):
    '''Require a host match.

    The host can either be anything ending
    with the EOL, a ; or a :, and optionally
    can be enclosed in [...] to allow
    containing : characters.
    '''
    match = host_re.match(uri)
    if match is None:
        raise ValueError('host must be specified')
    groups = match.groupdict()
    return groups['host'], groups['the_rest']


def parse_port(uri):
    '''Attempt to match a : and port number.'''
    match = port_re.match(uri)
    if match is None:
        return None, uri
    groups = match.groupdict()

    try:
        port = int(groups['port'])
    except ValueError:
        raise ValueError('port must be an integer')

    return port, groups['the_rest']


def parse_parameters(uri, post_params=None):
    '''Attempt to match many ;key=val pairs.

    If post_params is not None it will also be
    parsed and used to populate the parameters.
    '''
    the_rest = uri
    params = {}

    def parse_all(the_rest):
        match = param_re.match(the_rest)
        while match:
            groups = match.groupdict()
            if params.get(groups['key']):
                raise ValueError('parameters must be unique')
            params[groups['key']] = groups['val']
            the_rest = groups['the_rest']
            match = param_re.match(the_rest)
        return the_rest

    the_rest = parse_all(the_rest)
    if post_params:
        the_rest_post = parse_all(post_params)
        if the_rest_post:
            raise ValueError(f'could not parse {the_rest_post}')

    return params, the_rest


def parse_headers(uri):
    '''Attempt to match a ?key=val pair, then
    attempt to match many &key=val pairs.
    '''
    the_rest = uri
    headers = multidict.MultiDict()

    match = first_header_re.match(the_rest)
    while match:
        groups = match.groupdict()
        headers.add(groups['key'], groups['val'])
        the_rest = groups['the_rest']
        match = nth_header_re.match(the_rest)

    return headers, the_rest



'''URI object for parsing/building sip URIs with optional contact.'''
import copy
import random
import string
from typing import Optional
from multidict import MultiDict
from .parsing import (
    parse_post_params,
    parse_contact,
    parse_scheme,
    parse_user_part,
    parse_host,
    parse_port,
    parse_parameters,
    parse_headers,
)


def generate_tag():
    return ''.join([random.choice(string.hexdigits) for n in range(16)])


class URI:
    __slots__ = (
        '_contact',
        '_scheme',
        '_user',
        '_password',
        '_host',
        '_port',
        '_parameters',
        '_headers',
    )

    contact = property(lambda self: self._contact)
    scheme = property(lambda self: self._scheme)
    user = property(lambda self: self._user)
    password = property(lambda self: self._password)
    host = property(lambda self: self._host)
    port = property(lambda self: self._port)
    parameters = property(lambda self: self._parameters)
    headers = property(lambda self: self._headers)
    transport = property(lambda self: self.parameters.get('transport'))
    tag = property(lambda self: self.parameters.get('tag'))

    def __init__(self, uri: str):
        '''Parse a uri string into a URI.'''
        the_rest = uri
        post_params, the_rest = parse_post_params(the_rest)
        self._contact, the_rest = parse_contact(the_rest)
        self._scheme, the_rest = parse_scheme(the_rest)
        self._user, self._password, the_rest = parse_user_part(the_rest)
        self._host, the_rest = parse_host(the_rest)
        self._port, the_rest = parse_port(the_rest)
        self._parameters, the_rest = parse_parameters(the_rest, post_params)
        self._headers, the_rest = parse_headers(the_rest)
        if len(the_rest) > 0:
            raise ValueError(f'cannot parse {uri}')
        self._validate_and_default()

    @classmethod
    def build(cls, *,
              contact: Optional[str] = None,
              scheme: str,
              user: Optional[str]=None,
              password: Optional[str]=None,
              host: str,
              port: Optional[int]=None,
              parameters: Optional[dict]=None,
              headers: Optional[MultiDict]=None,
              ):
        '''Build a URI from kwargs for each component.'''
        self = object.__new__(cls)
        self._contact = contact
        self._scheme = scheme
        self._user = user
        self._password = password
        self._host = host
        self._port = port
        self._parameters = parameters if parameters else {}
        self._headers = headers if headers else MultiDict()
        self._validate_and_default()
        return self

    def with_contact(self, contact): return self._with(contact=contact)

    def with_scheme(self, scheme): return self._with(scheme=scheme)

    def with_user(self, user):
        if user and ':' in user[1:-1]:
            try:
                new_user, new_pass = user.split(':')
            except ValueError:
                raise ValueError('user must be either `user` or `user:pass`')
            return self._with(user=new_user, password=new_pass)
        elif user is None and self.password is not None:
            return self._with(user=user, password=None)
        else:
            return self._with(user=user)

    def with_password(self, password):
        if self.user is None:
            raise ValueError('cannot have a password without a user')
        return self._with(password=password)

    def with_host(self, host): return self._with(host=host)

    def with_port(self, port): return self._with(port=port)

    def with_parameters(self, parameters):
        for k, v in parameters.items():
            if v is None:
                parameters[k] = ''
        return self._with(parameters=parameters)

    def with_headers(self, headers):
        for k, v in headers.items():
            if v is None:
                headers[k] = ''
        return self._with(headers=headers)

    def with_transport(self, transport): return self._with(transport=transport)

    def with_tag(self, tag=None):
        if tag is None:
            if self.tag is not None:
                return self
            tag = generate_tag()
        return self._with(tag=tag)

    def _with(self, **kwargs):
        new = self.__deepcopy__()
        for attr in self.__slots__:
            name = attr[1:]
            if name in kwargs:
                setattr(new, attr, kwargs[name])
        for parameter in ['transport', 'tag']:
            if parameter in kwargs:
                new.parameters[parameter] = kwargs[parameter]
        new._validate_and_default()
        return new

    def _validate_and_default(self):
        '''Populate default port/transport if not explicitly
        given by uri string / build args, and ensure port is
        a valid value.'''
        if self.scheme not in ('sip', 'sips'):
            raise ValueError('scheme must be `sip` or `sips`')
        if self.port is None:
            self._port = 5061 if self.scheme == 'sips' else 5060
        if self.port not in range(1, 2**16):
            raise ValueError(f'port number {self.port} is not valid')
        if self.parameters.get('transport') is None:
            transport = 'tcp' if self.scheme == 'sips' else 'udp'
            self.parameters['transport'] = transport
        else:
            if self.scheme == 'sips' and self.transport == 'udp':
                raise ValueError(f'sips uris *must* use reliable transports,'
                                 f' {self.transport} is not reliable.')

    def __deepcopy__(self):
        return URI.build(
            contact=self.contact,
            scheme=self.scheme,
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            parameters=copy.deepcopy(self.parameters),
            headers=copy.deepcopy(self.headers),
        )

    def __str__(self):
        '''Give a normalized string representation of the URI.

        Normalization here means applying these rules:
            * only place uri in <> brackets if the contact is defined
            * always explicitly specify the port
            * always explicitly specify the transport parameter
        '''
        if self.user and self.password:
            user = f'{self.user}:{self.password}@'
        elif self.user:
            user = f'{self.user}@'
        else:
            user = ''

        parameter_pairs = ';'.join(['='.join([k, v])
                                    for k, v in self.parameters.items()])
        parameters = f';{parameter_pairs}'
        header_pairs = '&'.join(['='.join([k, v])
                                 for k, v in self.headers.items()])
        headers = f'?{header_pairs}' if len(header_pairs) else ''

        uri = (f'{self.scheme}:{user}{self.host}'
               f':{self.port}{parameters}{headers}')

        if self.contact:
            return f'"{self.contact}" <{uri}>'
        else:
            return uri

    def __eq__(self, other):
        # TODO: are the parameters/headers order-sensitive or not?
        return str(self) == str(other)

    def __hash__(self):
        # TODO: are the parameters/headers order-sensitive or not?
        return hash(str(self))



import pytest
from multidict import MultiDict
from ursine import URI


@pytest.mark.parametrize('uri', [
    'sip:localhost:port',
    'sip:localhost:0',
    'sip:localhost:70000',
    'sip:localhost?',
    'sip:localhost;',
    'sip:localhost&',
])
def test_invalid(uri):
    with pytest.raises(ValueError):
        URI(uri)


@pytest.mark.parametrize('uri,expect', [
    ('sip:localhost', 'sip:localhost:5060;transport=udp'),
    ('sips:localhost', 'sips:localhost:5061;transport=tcp'),
    ('<sip:localhost>', 'sip:localhost:5060;transport=udp'),
    (
        'John Doe <sip:localhost:5080?x=y&a=b>',
        '"John Doe" <sip:localhost:5080;transport=udp?x=y&a=b>',
    )
])
def test_to_str(uri, expect):
    assert str(URI(uri)) == expect


@pytest.mark.parametrize('uri1,uri2', [
    ('sip:localhost', 'sip:localhost'),
    ('sip:localhost', 'sip:localhost;transport=udp'),
    ('<sip:localhost>', 'sip:localhost'),
    ('Alice <sip:localhost>', '"Alice" <sip:localhost>'),
    ('SIP:localhost', 'sip:localhost'),
    ('<sip:localhost>;tag=foo', 'sip:localhost;tag=foo'),
    ('<sip:localhost> ;tag=foo', 'sip:localhost;tag=foo'),
    # TODO: test different parameter/header orderings?
    #       don't know if either/both are order sensitive
])
def test_equality(uri1, uri2):
    assert URI(uri1) == URI(uri2)


@pytest.mark.parametrize('uri1,uri2', [
    ('sip:localhost', 'sips:localhost'),
    ('Bob <sip:localhost>', 'sip:localhost'),
    ('Alice <sip:localhost>', 'alice <sip:localhost>'),
    ('sip:remotehost', 'sip:localhost')
])
def test_inequality(uri1, uri2):
    assert URI(uri1) != URI(uri2)


@pytest.mark.parametrize('kwargs,expect', [
    ({'scheme': 'sip', 'host': 'localhost'}, 'sip:localhost'),
    (
        {
            'scheme': 'sip', 'host': 'localhost',
            'parameters': {'transport': 'tcp'},
        },
        'sip:localhost;transport=tcp',
    ),
    (
        {
            'scheme': 'sips', 'host': '[::1]', 'port': 5080,
            'parameters': {'maddr': '[::dead:beef]'},
            'headers': MultiDict({'x': 'y', 'a': ''}),
        },
        'sips:[::1]:5080;maddr=[::dead:beef]?x=y&a=',
    ),
])
def test_build(kwargs, expect):
    assert URI.build(**kwargs) == URI(expect)


@pytest.mark.parametrize('original,attr,new,expect', [
    ('sip:localhost', 'user', 'jdoe', 'sip:jdoe@localhost'),
    ('sip:localhost;transport=tcp', 'scheme', 'sips', 'sips:localhost:5060'),
    ('sip:localhost', 'port', 5080, 'sip:localhost:5080'),
    ('sip:jdoe@localhost', 'user', None, 'sip:localhost'),
    ('"Mark" <sip:localhost>', 'contact', None, 'sip:localhost'),
    ('sip:user:pass@localhost', 'user', None, 'sip:localhost'),
    ('sip:localhost', 'user', 'user:pass', 'sip:user:pass@localhost'),
    ('sip:alice@localhost', 'password', 'pass', 'sip:alice:pass@localhost'),
    ('sip:localhost', 'transport', 'tcp', 'sip:localhost;transport=tcp'),
    ('sip:localhost', 'tag', 'bler', 'sip:localhost;transport=udp;tag=bler'),
    (
        'sip:localhost', 'parameters',
        {'maddr': '[::1]', 'foo': 'bar', 'x': ''},
        'sip:localhost;maddr=[::1];foo=bar;x=',
    ),
    (
        'sip:localhost', 'headers',
        {'ahhhh': '', 'foo': 'bar'},
        'sip:localhost?ahhhh=&foo=bar',
    ),
    (
        'sip:localhost', 'parameters',
        {'maddr': '[::1]', 'foo': 'bar', 'x': None},
        'sip:localhost;maddr=[::1];foo=bar;x=',
    ),
    (
        'sip:localhost', 'headers',
        {'ahhhh': None, 'foo': 'bar'},
        'sip:localhost?ahhhh=&foo=bar',
    ),
])
def test_modified_uri_creation(original, attr, new, expect):
    old_uri = URI(original)
    new_uri = getattr(old_uri, f'with_{attr}')(new)
    assert new_uri == URI(expect)
    assert new_uri != old_uri


@pytest.mark.parametrize('uri,tag,expect', [
    ('sip:localhost', None, None),
    ('sip:localhost', '5654', '5654'),
    ('sip:localhost;tag=2000', '5654', '5654'),
    ('sip:localhost;tag=2ace', None, '2ace'),
])
def test_tag_generation(uri, tag, expect):
    new_tag = URI(uri).with_tag(tag).tag
    if expect is None:
        assert new_tag
    else:
        assert new_tag == expect



#!/usr/bin/env python3
import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand


install_requires = [
    'multidict',
]

tests_require = [
    'pytest',
]


def readme():
    with open('README.md') as fp:
        return fp.read()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(name='ursine',
      version='0.2.2',
      description='library for SIP url handling/maninupation',
      long_description=readme(),
      long_description_content_type='text/markdown',
      author='Terry Kerr',
      author_email='t@xnr.ca',
      license='Apache 2',
      packages=['ursine'],
      keywords=['sip', 'voip', 'url'],
      url='https://github.com/sangoma/ursine',
      install_requires=install_requires,
      python_requires='>=3.6',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Programming Language :: Python :: 3 :: Only',
          'Programming Language :: Python :: 3.6',
          'Topic :: Communications :: Telephony',
          'Topic :: Communications :: Internet Phone',
          'License :: OSI Approved :: Apache Software License',
      ],
      tests_require=tests_require,
      cmdclass={'test': PyTest},
      )



import pytest
from ursine.parsing import (
    parse_contact,
    parse_scheme,
    parse_user_part,
    parse_host,
    parse_port,
    parse_parameters,
    parse_headers,
)


@pytest.mark.parametrize('uri,expect_contact,expect_the_rest', [
    ('sip:host', None, 'sip:host'),
    ('<sip:host>', None, 'sip:host'),
    ('John <sip:host>', 'John', 'sip:host'),
    ('"Mary" <sip:host>', 'Mary', 'sip:host'),
    ('"Alice Cooper" <sip:host>', 'Alice Cooper', 'sip:host'),
    ('"<>>:<<>" <sip:host>', '<>>:<<>', 'sip:host'),
    ('<sip:host', None, '<sip:host'),
    ('sip:host>', None, 'sip:host>'),
    ('Mary" <sip:host>', None, 'Mary" <sip:host>'),
    ('Alice sip:host', None, 'Alice sip:host'),
    (' sip:host', None, ' sip:host'),
])
def test_contact(uri, expect_contact, expect_the_rest):
    contact, the_rest = parse_contact(uri)
    assert contact == expect_contact
    assert the_rest == expect_the_rest


@pytest.mark.parametrize('uri,expect_scheme,expect_the_rest', [
    ('sip:', 'sip', ''),
    ('sips:', 'sips', ''),
    ('sip:localhost:5060', 'sip', 'localhost:5060'),
])
def test_scheme(uri, expect_scheme, expect_the_rest):
    scheme, the_rest = parse_scheme(uri)
    assert scheme == expect_scheme
    assert the_rest == expect_the_rest


@pytest.mark.parametrize('uri', [
    '', 'sisp:', ':', 'sip', 'sips', 'localhost:5060',
])
def test_scheme_fails(uri):
    with pytest.raises(ValueError):
        parse_scheme(uri)


@pytest.mark.parametrize('uri,expect_user,expect_pass,expect_the_rest', [
    ('bob:pass@some.tld', 'bob', 'pass', 'some.tld'),
    ('bob@some.tld', 'bob', None, 'some.tld'),
    ('some.tld', None, None, 'some.tld'),
    ('a@some.tld', 'a', None, 'some.tld'),
    ('@some.tld', None, None, '@some.tld'),
    (':some.tld', None, None, ':some.tld'),
])
def test_user_part(uri, expect_user, expect_pass, expect_the_rest):
    user, password, the_rest = parse_user_part(uri)
    assert user == expect_user
    assert password == expect_pass
    assert the_rest == expect_the_rest


@pytest.mark.parametrize('uri,expect_host,expect_the_rest', [
    ('some.tld', 'some.tld', ''),
    ('some.tld:5060;x=y', 'some.tld', ':5060;x=y'),
    ('tld', 'tld', ''),
    ('tld:5060;x=y', 'tld', ':5060;x=y'),
    ('[::1]:5060', '[::1]', ':5060'),
])
def test_host(uri, expect_host, expect_the_rest):
    host, the_rest = parse_host(uri)
    assert host == expect_host
    assert the_rest == expect_the_rest


@pytest.mark.parametrize('uri,expect_port,expect_the_rest', [
    (':5060', 5060, ''),
    (':1', 1, ''),
    (':5060;x=y', 5060, ';x=y'),
    (':5060?x=y', 5060, '?x=y'),
    (':65535', 65535, ''),
])
def test_port(uri, expect_port, expect_the_rest):
    port, the_rest = parse_port(uri)
    assert port == expect_port
    assert the_rest == expect_the_rest


@pytest.mark.parametrize('uri,expect_params,expect_the_rest', [
    (';ten=10', {'ten': '10'}, ''),
    (';ten=10;five=5', {'ten': '10', 'five': '5'}, ''),
    (';ten=10;five=5?x=y', {'ten': '10', 'five': '5'}, '?x=y'),
])
def test_parameters(uri, expect_params, expect_the_rest):
    params, the_rest = parse_parameters(uri)
    assert params == expect_params
    assert the_rest == expect_the_rest


@pytest.mark.parametrize('uri,expect_headers,expect_the_rest', [
    ('?x=y', {'x': ['y']}, ''),
    ('?x=y&x=z', {'x': ['y', 'z']}, ''),
    ('?x=y&a=b', {'x': ['y'], 'a': ['b']}, ''),
    ('?x=&y=', {'x': [''], 'y': ['']}, ''),
])
def test_headers(uri, expect_headers, expect_the_rest):
    headers, the_rest = parse_headers(uri)
    for k, v in expect_headers.items():
        assert headers.popall(k) == v
    assert len(headers) == 0
    assert the_rest == expect_the_rest
